// $Id: testing_io.c,v 1.8 2017/04/10 12:50:39 leavens Exp leavens $
// Author: Gary T. Leavens with help from Lawrence Kirby's post
//         at http://c-faq.com/stdio/rd.kirby.c about redirecting stdout.
// Some helpers for testing interactive I/O programs and procedures

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "tap.h"
#include "testing_io.h"

#define BUFSIZE 512

/* requires: size of the target is at least the length
 *           of str1 + length of str2 + 1 and the arguments are disjoint
 * modifies: target[*]
 * ensures: target is assigned the concatenation of str1 followed by str2. */
static
void strconcat(char *target, char *str1, char *str2, size_t size) {
    size_t len1 = strlen(str1);
    size_t len2 = strlen(str2);
    if (size >= len1+len2+1) {
	sprintf(target, "%s%s", str1, str2);
    } else {
	bail_out(1, "not enough space for strconcat");
    }
}

/* copy src to target, taking out \r chars.
   Requires, target has at least as much space as there are chars in src. */
static
void nocrs(char *target, char *src) {
    // all the chars in old(src) up to but not including *src are in old(target)
    while (*src != '\0') {
	if (*src != '\r') {
	    *target = *src;
	    target++;
	}
	src++;
    }
    // *src == '\0'
    *target = '\0';
}

// check to see if an open worked
// issue error messages and bail if not.
static
void check_open(FILE *f, char *fn) {
    if (f == NULL || ferror(f)) {
	perror("ERROR");
	bail_out(2, "can't read file \"%s\"\n", fn);
    }
}

// fix the buffer buf by putting a null character at the element cs.
// Requires *buf is allocated and has size BUFSIZE.
// Checks for errors in the file f named fn and prints error messages if needed
static
void fixup(char *buf, size_t cs) {
    if (cs < BUFSIZE-1) {
	buf[cs] = '\0';
    }
}

// read all the characters in f into buf, which has size sz,
// Requires: sz > 0 and f is open for reading;
static
size_t freadchars(char *buf, size_t sz, FILE *f) {
    buf[0] = '\0';
    size_t i = 0;
    char c = getc(f);
    // the chars in f before the ith position are in buf[0:i]
    // and i < sz
    while (i < sz-1 && c != -1) {
	buf[i] = c;
	c = getc(f);
	i++;
    }
    // the chars in f before the ith position are in buf[0:i] and i < sz
    buf[i] = '\0';
    return i;
}

/* compare the output in prefix.txt
 * with the expected output in prefix.expected,
 * and cause an assertion violation if they are not the same. */
static
void compare(char *outfilename, char *prefix) {
    // read in the actual output and the expectd output
    FILE *outf = fopen(outfilename, "r");
    check_open(outf, outfilename);
    char expectedfilename[BUFSIZE];

    size_t cs;
    char readout_contents[BUFSIZE];
    cs = freadchars(readout_contents, BUFSIZE, outf);
    fixup(readout_contents, cs);
    char read_nocrs[BUFSIZE];
    nocrs(read_nocrs, readout_contents);

    char exp_contents[BUFSIZE];
    strconcat(expectedfilename, prefix, ".expected", BUFSIZE);
    FILE *expf = fopen(expectedfilename, "r");
    check_open(expf, expectedfilename);
    cs = freadchars(exp_contents, BUFSIZE, expf);
    fixup(exp_contents, cs);
    char exp_nocrs[BUFSIZE];
    nocrs(exp_nocrs, exp_contents);

    // they should be the same
    is(read_nocrs, exp_nocrs,
       "output should be expected output for %s.in", prefix);
    fclose(outf);
    fclose(expf);
}

/* redirect the given stream to the given file name, 
   which is opened in the given mode. 
   A duplicate of the old file descriptor is returned. */
static
int redirect_from(FILE *stream, char* name, char * mode, fpos_t *pos) {
    if (strcmp(mode, "w") == 0) {
	fflush(stream);
    }
    fgetpos(stream, pos);
    int saved_fd = dup(fileno(stream));
    freopen(name, mode, stream);
    return saved_fd;
}

/* redirect the given (saved) file descriptor to the given stream
   which was opened in the given mode. */
static
void redirect_back(int saved_fd, FILE *stream, char *mode, const fpos_t *pos) {
    if (strcmp(mode, "w") == 0) {
	fflush(stream);
    }
    dup2(saved_fd, fileno(stream));
    close(saved_fd);
    clearerr(stream);
    fsetpos(stream, pos);   // for C99
}

/* Run the program in the file named progn (which must be in the PATH)
 * with input from file '${prefix}.in'
 * and output going to file '${prefix}.txt'.
 * The result is the return code from running progn. */
int testprog(char *progn, char *prefix) {
    // create the input and output file names
    char infilename[BUFSIZE];
    strconcat(infilename, prefix, ".in", BUFSIZE);
    char outfilename[BUFSIZE];
    strconcat(outfilename, prefix, ".txt", BUFSIZE);
    // run the procedure proc
    // with input from infilename and output to outfilename
    int ret = runprog(progn, infilename, outfilename);
    compare(outfilename, prefix);
    return ret;
}

/* Run the 0-argument procedure proc with input from file named infn
 * (which acts as stdin) and output going to the file named outfn
 * (which captures stdout). */
int runprog(char *progn, char *infn, char *outfn) {
    // replace the standard streams
    char command[BUFSIZE];
    sprintf(command, "./%s < %s > %s 2>&1", progn, infn, outfn);
    return system(command);
}

/* Run the 0-argument procedure proc with input from file '${prefix}.in'
 *  and output going to file '${prefix}.txt'. */
int testproc(int (*proc)(), char *prefix) {
    // create the input and output file names
    char infilename[BUFSIZE];
    strconcat(infilename, prefix, ".in", BUFSIZE);
    char outfilename[BUFSIZE];
    strconcat(outfilename, prefix, ".txt", BUFSIZE);
    // run the procedure proc
    // with input from infilename and output to outfilename
    int ret = runproc(proc, infilename, outfilename);
    compare(outfilename, prefix);
    return ret;
}

/* Run the 0-argument procedure proc with input from file named infn
 * (which acts as stdin) and output going to the file named outfn
 * (which captures stdout). */
int runproc(int (*proc)(), char *infn, char *outfn) {
    // replace the standard streams
    fpos_t in_pos;
    int saved_stdin = redirect_from(stdin, infn, "r", &in_pos);
    fpos_t out_pos;
    int saved_stdout = redirect_from(stdout, outfn, "w", &out_pos);

    int ret = (*proc)();

    // undo the replacements
    redirect_back(saved_stdout, stdout, "w", &out_pos);
    redirect_back(saved_stdin, stdin, "r", &in_pos);

    return ret;
}
